import 'package:flutter/widgets.dart';
import 'package:provider/provider.dart';

import 'observable_model.dart';
import 'subscriber_gadget.dart';

/// A function used to initialize the internal state of
/// a given a view-model object [T].
typedef InitViewModel<T> = void Function(T viewModel);

/// A function used to dispose the internal state of
/// a given a view-model object [T].
typedef DisposeViewModel<T> = void Function(T viewModel);

/// A function used to build a Widget given a view-model object [T].
typedef ViewModelGadgetBuilder<T> = Widget Function(
  BuildContext context,
  T viewModel,
);

/// A Gadget that encapsulates and handles the state of a view model.
///
/// During initialization, the view-model object [T] is automatically registered
/// into the [ObservableModel] and the optional [init] callback is executed.
/// 
/// If the [subscribeToViewModel] property is true (default value), the [ViewModelGadget]
/// registers itself as a subscriber for the specified view-model: in this way, 
/// eventual changes to the view-model object can trigger the[builder] function,
/// effectively updating the UI.
/// 
/// ```dart
/// ViewModelGadget<MyViewModel>(
///   viewModel: MyViewModel(),
///   builder: (context, viewModel) => Column(
///     children: [
///       Text('${viewModel.someText}'),
///       Text('${viewModel.otherText}'),
///     ]
///   )
/// )
/// ```
/// 
/// Otherwise, [SubscriberGadget]s can be used in order to register some specific widgets
/// to the view-model (or some properties of it).
/// 
/// ```dart
/// ViewModelGadget<MyViewModel>(
///   viewModel: MyViewModel(),
///   subscribeToViewModel: false,
///   builder: (context, viewModel) => Column(
///     children: [
///       SubscriberGadget<MyViewModel>(
///         propertiesSelector: (viewModel) => [viewModel.someText]
///         builder: (_, viewModel) => Text('${viewModel.someText}')
///       ),
///       SubscriberGadget<MyViewModel>(
///         propertiesSelector: (viewModel) => [viewModel.otherText]
///         builder: (_, viewModel) => Text('${viewModel.otherText}')
///       ),
///     ]
///   )
/// )
/// ```
///
/// When the widget is disposed, the view-model object [T] is automatically
/// removed from the [ObservableModel] and the optional [dispose] callback is executed.
/// The [dispose] callback can be used to manually dispose stateful components of
/// the given view-model object (like a [TextEditingController]).
class ViewModelGadget<T> extends StatefulWidget {
  const ViewModelGadget({
    Key? key,
    this.init,
    required this.builder,
    this.dispose,
    required this.viewModel,
    this.subscribeToViewModel = false,
  }) : super(key: key);

  final InitViewModel<T>? init;
  final DisposeViewModel<T>? dispose;
  final ViewModelGadgetBuilder<T> builder;
  final T viewModel;
  final bool subscribeToViewModel;

  @override
  _ViewModelGadgetState<T> createState() => _ViewModelGadgetState<T>();
}

class _ViewModelGadgetState<T> extends State<ViewModelGadget<T>> {
  late ObservableModel model;

  @override
  void initState() {
    super.initState();
    if (widget.init != null) {
      widget.init!(widget.viewModel);
    }
    model = context.read<ObservableModel>()
      ..putValue<T>(value: widget.viewModel);
  }

  @override
  void dispose() {
    if (widget.dispose != null) {
      widget.dispose!(model.getValue<T>()!);
    }
    model.removeValue<T>(closeSubject: true);
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    if (widget.subscribeToViewModel) {
      return SubscriberGadget<T>(
        builder: (context, viewModel) => widget.builder(context, viewModel),
      );
    }
    final model = context.read<ObservableModel>();
    return widget.builder(context, model.getValue<T>()!);
  }
}
